/* ========================================================================== */
/*! \file
 * \brief Message Digests
 *
 * Copyright (c) 2012-2014 by the developers. See the LICENSE file for details.
 *
 * If nothing else is specified, function return zero to indicate success
 * and a negative value to indicate an error.
 */


/* ========================================================================== */
/* Include headers */

#include "posix.h"  /* Include this first because of feature test macros */

#include "config.h"

#if CFG_USE_TLS  /* This requires data from "config.h" */
#  include <openssl/evp.h>
#endif  /* CFG_USE_TLS */
#include <string.h>

#include "digest.h"
#include "main.h"


/* ========================================================================== */
/*! \addtogroup DIGEST
 *
 * This module provides optional support for message digest algorithms. It is
 * intended to comply with RFC3174.
 *
 * \attention
 * This module currently works only if TLS support is available (because both
 * share the cryptographic functions provided by OpenSSL).
 */
/* @{ */


/* ========================================================================== */
/*! \brief Secure Hash Algorithm SHA-1
 *
 * \param[in]  text      Message
 * \param[in]  text_len  Message length
 * \param[out] md        Message digest
 *
 * The values of paramater \e text_len must represent bytes.
 *
 * On success, the result is written to \e md and has a length of
 * \ref DIGEST_SHA1_LEN bytes.
 *
 * \return
 * - 0 on success
 * - -1 on error
 */

int  digest_sha1(const char*  text, size_t  text_len, unsigned char*  md)
{
   int  res = -1;
#if CFG_USE_TLS
   EVP_MD_CTX*  ctx;
   unsigned int  md_len = 0;
   int  rv;

   /* Create context */
   ctx = EVP_MD_CTX_create();
   if(NULL == ctx)  { print_error("DIGEST: Creating context failed"); }
   /* Check whether length value can be represented as data type 'int' */
   else if(!(POSIX_INT_MAX >= text_len))
   {
      print_error("DIGEST: Message or key is too long");
   }
   else
   {
      /* Delegate job to OpenSSL */
      rv = EVP_DigestInit_ex(ctx, EVP_sha1(), NULL);
      if(1 == rv)
      {
         rv = EVP_DigestUpdate(ctx, (void*) text, text_len);
         if(1 == rv)
         {
            rv = EVP_DigestFinal_ex(ctx, md, &md_len);
            if(1 == rv && DIGEST_SHA1_LEN == (size_t) md_len)  { res = 0; }
         }
      }
   }
   /* Destroy context */
   if(NULL != ctx)  { EVP_MD_CTX_destroy(ctx); }
#endif  /* CFG_USE_TLS */

   return(res);
}


/* ========================================================================== */
/*! \brief Initialize message digest module */

void  digest_init(void)
{
   return;
}


/* ========================================================================== */
/*! \brief Shutdown message digest module */

void  digest_exit(void)
{
   return;
}


/* @} */

/* EOF */
