/* ========================================================================== */
/*! \file
 * \brief Article filtering
 *
 * Copyright (c) 2012-2015 by the developers. See the LICENSE file for details.
 *
 * If nothing else is specified, function return zero to indicate success
 * and a negative value to indicate an error.
 */


/* ========================================================================== */
/* Include headers */

#include "posix.h"  /* Include this first because of feature test macros */

#include <string.h>

#include "conf.h"
#include "core.h"
#include "encoding.h"
#include "filter.h"
#include "fileutils.h"
#include "main.h"


/* ========================================================================== */
/*! \addtogroup FILTER
 *
 * Location of scorefile: \c ~/.$CFG_NAME/scorefile
 *
 * Any line starting with \c # is treated as a comment (not parsed and ignored).
 * All other lines are parsed as entries with 3 colon-separated fields:
 * - Field 1: Type (indicating the target element and matching method)
 * - Field 2: Score (signed integer value)
 * - Field 3: String
 *
 * No whitespace is allowed between the fields and separators.
 */
/* @{ */


/* ========================================================================== */
/* Constants */

/*! \brief Permissions for score file */
#define FILTER_PERM  (posix_mode_t) (POSIX_S_IRUSR | POSIX_S_IWUSR)

/*! \name Score limits
 *
 * Type must be \c int .
 * Minimum and maximum values are \c INT_MIN and \c INT_MAX .
 */
/* @{ */
#define FILTER_SCORE_MAX  INT_MAX
#define FILTER_SCORE_MIN  INT_MIN
/* @} */


/* ========================================================================== */
/* Data types */

/* Locale 'LC_CTYPE' codeset */
enum filter_cs
{
   FILTER_CS_ASCII,
   FILTER_CS_ISO8859_1,
   FILTER_CS_UTF_8
};

/* Data types of score entries
 *
 * The IDs must start with value 0 (unknown type) and must be contiguous
 * Most used types should be defined first for better performance
 */
enum filter_entry_type
{
   SCORE_TYPE_UNKNOWN     = 0,
   /* ----------------------------------------------------------------------- */
   SCORE_TYPE_FROM        = 1,  /* Literal string vs. 'From' */
   SCORE_TYPE_FROM_ERE    = 2,  /* Extended regular expression vs. 'From' */
   SCORE_TYPE_SUBJECT     = 3,  /* Literal string vs. 'Subject' */
   SCORE_TYPE_SUBJECT_ERE = 4,  /* Extended regular expression vs. 'Subject' */
   SCORE_TYPE_MSGID_ERE   = 5,  /* Extended regular expr. vs. 'Message-ID' */
   SCORE_TYPE_GROUP       = 6,  /* Literal string vs. element of 'Newsgroups' */
   /* ----------------------------------------------------------------------- */
   SCORE_END_OF_LIST      = 7
};

/* Scoring rule */
struct filter
{
   enum filter_entry_type  type;
   int  value;
   const char*  string;
   int  found;
   struct filter*  next;
   enum filter_cs  cs;
#if CFG_USE_CLB || CFG_USE_XSI
   posix_regex_t*  ere;
#endif  /* CFG_USE_CLB || CFG_USE_XSI */
};


/* ========================================================================== */
/* Variables */

/* Rule names for first field of score file entries
 *
 * \attention
 * The data type 'enum filter_entry_type' must be suitable as index!
 */
static const char*  filter_type_name[] =
{
   "unknown",
   /* ----------------------------------------------------------------------- */
   "from",
   "from_ere",
   "subject",
   "subject_ere",
   "msgid_ere",
   "group",
   /* ----------------------------------------------------------------------- */
   "eol"
};

static enum filter_cs  filter_locale = FILTER_CS_ASCII;
#if CFG_USE_CLB || CFG_USE_XSI
static enum filter_cs  testgroup_cs;
static posix_regex_t*  testgroup_ere = NULL;
#endif  /* CFG_USE_CLB || CFG_USE_XSI */
static size_t  score_len_max = 1000;
static struct filter*  scores = NULL;
static const char  scorefile_name[] = "scorefile";


#if CFG_USE_CLB || CFG_USE_XSI


/* ========================================================================== */
/* Print error message if system failed to compile a regular expression
 *
 * \param[in] code  Error code
 * \param[in] ere   Pointer to compiled ERE
 *
 * The value \e code must be nonzero and the last error code returned from
 * \c regcomp() for \e ere according to the POSIX standard:
 * <br>
 * http://pubs.opengroup.org/onlinepubs/9699919799/functions/regcomp.html
 *
 * \attention
 * The definition "last error code returned" is incomplete (may mean the last
 * call from a thread or the last call from the whole process)
 * => Always use the filter module from the same, single thread.
 *
 * \note
 * The error message is always formatted using the POSIX locale.
 */

static void  filter_print_ere_error(int  code, posix_regex_t*  ere)
{
   const char*  mod_name = "FILTER: ";
   size_t  mod_len = strlen(mod_name);
   size_t  len;
   char*  buf = NULL;

   if(!code)
   {
      print_error("FILTER: Can't process invalid error code");
   }
   else
   {
#  if CFG_USE_XSI
      /* Don't use NLS for error messages on stderr */
      posix_setlocale(POSIX_LC_MESSAGES, "POSIX");
#  endif  /* CFG_USE_XSI */
      len = posix_regerror(code, ere, buf, 0);
      if(!len || POSIX_SIZE_MAX - mod_len < len)
      {
         print_error("FILTER: Error message has invalid size");
      }
      else
      {
         buf = (char*) posix_malloc(mod_len + len);
         if(NULL == buf)
         {
            print_error("FILTER: Cannot allocate memory for error message");
         }
         else
         {
            strncpy(buf, mod_name, mod_len);
            posix_regerror(code, ere, &buf[mod_len], len);
            print_error(buf);
            posix_free((void*) buf);
         }
      }
#  if CFG_USE_XSI
      posix_setlocale(POSIX_LC_MESSAGES, "");
#  endif  /* CFG_USE_XSI */
   }
}


/* ========================================================================== */
/* Compile ERE
 *
 * \param[out] cs      Pointer to codeset of character classification locale
 * \param[out] ere     Pointer to compiled ERE
 * \param[in]  string  Raw ERE pattern
 *
 * \note
 * On success, the caller is responsible for freeing the ressources allocated
 * for the object pointed to by \e ere .
 */

static int  filter_compile_ere(enum filter_cs*  cs, posix_regex_t**  ere,
                               const char*  string)
{
   int  res = 0;
   int  rv;
   enum enc_mime_cs  charset = ENC_CS_UNKNOWN;
   const char*  pat;
   const char*  p = NULL;

   *cs = filter_locale;
   if(FILTER_CS_ISO8859_1 == filter_locale)
   {
      /* Convert string to ISO8859-1 */
      p = enc_convert_to_8bit(&charset, string, NULL);
      if(NULL == p)  { res = -1; }
      else if(ENC_CS_ISO8859_1 != charset)  { res = -1; }
   }
   else if(FILTER_CS_UTF_8 != filter_locale)
   {
      /* Treat unsupported codeset as ASCII */
      *cs = FILTER_CS_ASCII;
      res = enc_ascii_check(string);
   }
   if(res)
   {
      /* String cannot be process without UTF-8 locale */
      print_error("FILTER: ERE cannot be used with current locale");
   }
   else
   {
      /* Allocate memory */
      *ere = (posix_regex_t*)  posix_malloc(sizeof(posix_regex_t));
      if(NULL == *ere)
      {
         print_error("FILTER: Cannot allocate memory for regular expression");
         res = -1;
      }
      else
      {
         /* Compile regular expression if required */
         pat = string;
         if(FILTER_CS_ISO8859_1 == filter_locale)  { pat = p; }
         rv = posix_regcomp(*ere, pat, POSIX_REG_EXTENDED | POSIX_REG_NOSUB);
         if(rv)
         {
            print_error("FILTER: Compiling regular expression failed");
            filter_print_ere_error(rv, *ere);
            posix_free((void*) *ere);
            res = -1;
         }
         else
         {
            printf("%s: FILTER: Compiling regular expression\n", CFG_NAME);
         }
      }
   }
   /* Release memory for ISO8859-1 string */
   if(NULL != p && string != p)  { enc_free((void*) p); }

   return(res);
}


#endif  /* CFG_USE_CLB || CFG_USE_XSI */


/* ========================================================================== */
/* Score rule constructor
 *
 * \param[out] new_rule  Pointer to new rule
 * \param[in]  type      Type of new rule
 * \param[in]  score     Score if rule matches
 * \param[in]  string    Literal string or regular expression
 * \param[in]  dcre      Don't compile regular expression if nonzero
 *
 * The parameter \e dcre should only be nonzero for exporting the rule data
 * back to the scorefile (for comparing the raw regular expression strings).
 *
 * \attention
 * The parameter \e string must point to a memory block allocated with the
 * function \ref posix_malloc() and will become part of the created object.
 * On error the caller stay responsible to free the memory for \e string .
 *
 * \note
 * On success, the caller is responsible for freeing the ressources allocated
 * for the object \e rule (use \ref filter_score_rule_destructor() function).
 *
 * On error \c NULL is written to \e new_rule .
 */

static int  filter_score_rule_constructor(struct filter**  new_rule,
                                          enum filter_entry_type  type,
                                          int  score, const char*  string,
                                          int  dcre)
{
   int   res = 0;

   /* Create new rule */
   *new_rule = (struct filter*) posix_malloc(sizeof(struct filter));
   if(NULL == *new_rule)
   {
      print_error("FILTER: Cannot allocate memory for score rule");
      res = -1;
   }
   else
   {
      (*new_rule)->type = type;
      (*new_rule)->value = score;
      (*new_rule)->string = string;
      (*new_rule)->found = 0;
      (*new_rule)->next = NULL;
      (*new_rule)->cs = FILTER_CS_UTF_8;
   }
#if CFG_USE_CLB || CFG_USE_XSI
   if(!res)
   {
      (*new_rule)->ere = NULL;
      /* Compile regular expression */
      if(!dcre)
      {
         switch((*new_rule)->type)
         {
            case SCORE_TYPE_FROM_ERE:
            case SCORE_TYPE_SUBJECT_ERE:
            case SCORE_TYPE_MSGID_ERE:
            {
               res = filter_compile_ere(&(*new_rule)->cs, &(*new_rule)->ere,
                                        string);
               break;
            }
            default:
            {
               /* Rule do not use a regular expression */
               break;
            }
         }
      }
   }
#endif  /* CFG_USE_CLB || CFG_USE_XSI */

   /* Check for error */
   if(res)
   {
      posix_free((void*) *new_rule);
      *new_rule = NULL;
   }

   return(res);
}


/* ========================================================================== */
/* Score rule destructor
 *
 * \param[in,out] rule  Object created by \ref filter_score_rule_contructor()
 */

static void  filter_score_rule_destructor(struct filter**  rule)
{
   if(NULL != rule)
   {
      posix_free((void*) (*rule)->string);
#if CFG_USE_CLB || CFG_USE_XSI
      if(NULL != (*rule)->ere)
      {
         /* Destroy regular expression object */
         posix_regfree((*rule)->ere);
         posix_free((void*) (*rule)->ere);
      }
#endif  /* CFG_USE_CLB || CFG_USE_XSI */
      posix_free((void*) *rule);
      *rule = NULL;
   }
}


/* ========================================================================== */
/* Calculate and clamp new score
 *
 * \param[in] val  Old value
 * \param[in] val  Difference to old value that should be aplied
 *
 * \return
 * - Updated score
 */

static int  filter_score_add(int  val, int  diff)
{
   /* Check for increase */
   if(0 < diff)
   {
      if(0 < val)
      {
         /* Clamp to upper limit */
         if(FILTER_SCORE_MAX - val < diff)  { val = FILTER_SCORE_MAX; }
         else  { val += diff; }
      }
      else  { val += diff; }
   }

   /* Check for decrease */
   if(0 > diff)
   {
      if(0 > val)
      {
         /* Clamp to lower limit */
         if(FILTER_SCORE_MIN - val > diff)  { val = FILTER_SCORE_MIN; }
         else  { val += diff; }
      }
      else  { val += diff; }
   }

   return(val);
}


/* ========================================================================== */
/* Add score rule
 *
 * \param[in] new_rule  Rule to add
 */

static void  filter_add_score_rule(struct filter*  new_rule)
{
   struct filter*  last_rule = scores;

   if(NULL == last_rule)  { scores = new_rule; }
   else
   {
      /* Append rule to end of list */
      while(NULL != last_rule->next)  { last_rule = last_rule->next; }
      last_rule->next = new_rule;
   }
}


/* ========================================================================== */
/* Delete scoring rules */

static void  filter_delete_score_rules(void)
{
   struct filter*  rule = scores;
   struct filter*  next_rule;

   while(NULL != rule)
   {
      next_rule = rule->next;
      filter_score_rule_destructor(&rule);
      rule = next_rule;
   }
   scores = NULL;
}


/* ========================================================================== */
/* Decode score rule from score file
 *
 * \param[out] rule  Pointer to decoded rule object
 * \param[in]  line  Line from score file
 * \param[in]  len   Line buffer size
 * \param[in]  dcre  Dont't compile potential regular expression if nonzero
 *
 * \note
 * On success, the caller is responsible for freeing the ressources allocated
 * for the object \e rule (use \ref filter_score_rule_destructor() function).
 */

static int  filter_decode_rule(struct filter**  rule, char*  line, size_t  len,
                               int  dcre)
{
   int  res = -1;
   const char*  p;
   enum filter_entry_type  type;
   size_t  start;
   int  score = 0;
   char*  string = NULL;
   int  rv;

   /* Init rule pointer to defined value */
   *rule = NULL;

   /* Get type */
   type = SCORE_TYPE_UNKNOWN;
   while(SCORE_END_OF_LIST != ++type)
   {
      p = filter_type_name[type];
      start = strlen(p);
      if(start < len)
      {
         if(!strncmp(line, p, start))
         {
            /* Verify that not only the first part has matched */
            if(':' == line[start])  { break; }
         }
      }
   }
   if(SCORE_END_OF_LIST == type)  { type = SCORE_TYPE_UNKNOWN; }

   /* Allocate memory for data field */
   string = (char*) posix_malloc((size_t) len);
   if(NULL == string)
   {
      print_error("FILTER: Cannot allocate memory for score rule parser");
   }
   else
   {
      /* Decode data */
      switch(type)
      {
         case SCORE_TYPE_FROM:
         case SCORE_TYPE_FROM_ERE:
         case SCORE_TYPE_SUBJECT:
         case SCORE_TYPE_SUBJECT_ERE:
         case SCORE_TYPE_MSGID_ERE:
         case SCORE_TYPE_GROUP:
         {
            rv = sscanf(&line[start], ":%d:%[^\n]", &score, string);
            if(2 != rv)
            {
               print_error("FILTER: Invalid entry in score file");
            }
            else
            {
               /* Success */
               res = 0;
            }
            break;
         }
         default:
         {
            print_error("FILTER: Unknown entry type in score file");
            break;
         }
      }
   }

   /* Create score rule object */
   if(!res)
   {
#if 0
      /* For debugging */
      printf("=============================\n");
      printf("   type  : %d\n", type);
      printf("   score : %d\n", score);
      printf("   string: %s\n", string);
#endif
      res = filter_score_rule_constructor(rule, type, score, string, dcre);
   }

   /* Check for error */
   if(res)  { posix_free((void*) string); }

   return(res);
}


/* ========================================================================== */
/* Encode score rule for score file
 *
 * \param[in,out] line  Pointer to line for score file
 * \param[in,out] len   Pointer to line buffer size
 * \param[in]     rule  Rule to encode
 *
 * \attention
 * The pointer \e line must be \c NULL or point to a dynamically allocated
 * buffer.
 */

static int  filter_encode_rule(char**  line, size_t*  len, struct filter*  rule)
{
   const char*  frt = NULL;  /* Filter Rule Type */
   int  res = -1;
   int  rv;
   char*  p = NULL;
   size_t  l = 0;

   /* Create new score entry */
   switch(rule->type)
   {
      case SCORE_TYPE_FROM:
      case SCORE_TYPE_FROM_ERE:
      case SCORE_TYPE_SUBJECT:
      case SCORE_TYPE_SUBJECT_ERE:
      case SCORE_TYPE_MSGID_ERE:
      case SCORE_TYPE_GROUP:
      {
         frt = filter_type_name[rule->type];
         /* Calculate buffer size */
         l = strlen(frt);     /* Type ID */
         l += (size_t) 1;     /* Field separator */
         l += score_len_max;  /* Score value */
         l += (size_t) 1;     /* Field separator */
         l += strlen(rule->string);
         l += (size_t) 1;     /* LF line termination */
         l += (size_t) 1;     /* NUL termination */
         p = (char*) posix_malloc(l);
         if(NULL == p)
         {
            print_error("FILTER: Cannot allocate memory for score entry");
         }
         else
         {
            rv = posix_snprintf(p, l, "%s:%d:%s\n",
                                frt, rule->value, rule->string);
            if(0 > rv || (size_t) rv >= l)
            {
               print_error("FILTER: Encoding score entry failed");
               posix_free((void*) p);
            }
            else
            {
               /* Success => Replace line buffer */
               posix_free((void*) *line);
               *line = p;
               *len = l;
               res = 0;
            }
         }
         break;
      }
      default:
      {
         print_error("FILTER: Encoding unknown entry type failed");
         break;
      }
   }

   return(res);
}


/* ========================================================================== */
/* Check whether score rule match
 *
 * \param[in] line  Line from score file
 * \param[in] len   Line buffer size
 * \param[in] rule  Rule to match
 */

static int  filter_check_entry(char*  line, size_t  len, struct filter*  rule)
{
   int  res = -1;
   struct filter*  current_rule;

   /* Decode line */
   if('#' != line[0])
   {
      /* Set 'dcre' flag, we only want to compare the strings of the rule */
      if(!filter_decode_rule(&current_rule, line, len, 1))
      {
         /* Check whether score has changed */
         if(current_rule->type == rule->type)
         {
            if(!strcmp(current_rule->string, rule->string))
            {
               /* Match detected */
               res = 0;
            }
         }
         /* Destroy current rule object */
         filter_score_rule_destructor(&current_rule);
      }
   }

   return(res);
}


/* ========================================================================== */
/* Export score rules
 *
 * \param[in] fs      Stream corresponding to old configuration
 * \param[in] fs_tmp  Stream corresponding to new configuration
 *
 * The current data in memory is merged with the data from \e fs and written
 * to \e fs_tmp .
 */

static int  filter_export_score_rules(FILE*  fs, FILE*  fs_tmp)
{
   int  res = -1;
   char*  line = NULL;
   size_t  len = 0;
   posix_ssize_t  readlen;
   int  rv;
   struct filter*  rule;

   while(1)
   {
      /* Read line */
      readlen = posix_getline(&line, &len, fs);
      if(-1 == readlen)
      {
         if(POSIX_ENOMEM == posix_errno)
         {
            print_error("FILTER: Cannot assign memory for score file parser");
         }
         else
         {
            /* Check for error */
            if(ferror(fs))
            {
               print_error("FILTER: Parse error in score file");
            }
            /* Check for EOF */
            else if(feof(fs))
            {
               res = 0;
            }
         }
      }
      if(0 >= readlen)  { break; }
      else
      {
         /* Update data */
         rule = scores;
         while(NULL != rule)
         {
            if(!rule->found && !filter_check_entry(line, len, rule))
            {
               /* Match => Update */
               rule->found = 1;
               filter_encode_rule(&line, &len, rule);
               break;
            }
            rule = rule->next;
         }

         /* Write line to new config file */
         rv = fprintf(fs_tmp, "%s", line);
         if(0 > rv)  { break; }
      }
   }

   /* Add missing entries to end of config file */
   if(!res)
   {
      rule = scores;
      while(NULL != rule)
      {
         if(!rule->found)
         {
            rv = filter_encode_rule(&line, &len, rule);
            if(rv)  { res = -1;  break; }
            else
            {
               /* Write new entry */
               rv = fprintf(fs_tmp, "%s", line);
               if(0 > rv)
               {
                  res = -1;
                  break;
               }
            }
         }
         rule = rule->next;
      }
   }

   /* Release memory for line buffer */
   posix_free((void*) line);

   return(res);
}


/* ========================================================================== */
/* Get scorefile pathname
 *
 * This function must be thread safe.
 * The caller is responsible to free the memory for the buffer on success.
 */

static int  filter_get_pathname(char**  pathname, const char*  filename)
{
   int  res = -1;
   const char*  homedir = NULL;
   const char  confdir[] = "/." CFG_NAME "/";
   int  rv;

   /* Get home directory from environment */
   rv = ts_getenv("HOME", &homedir);
   if (!rv)
   {
      /* Check home directory path */
      rv = fu_check_homedir(homedir);
      /* Create score file pathname */
      if (!rv)
      {
         /* The additional byte is for the terminating 0 */
         *pathname = (char*) posix_malloc(strlen(homedir) + strlen(confdir)
                                          + strlen(filename) + (size_t) 1);
         if (NULL == *pathname)
         {
            print_error(
               "FILTER: Cannot allocate memory for score file pathname");
         }
         else
         {
            /* Create configuration directory if it doesn't exist */
            strcpy(*pathname, homedir);
            strcat(*pathname, confdir);
            rv = posix_mkdir(*pathname, (posix_mode_t) POSIX_S_IRWXU);
            if (!rv || (-1 == rv && POSIX_EEXIST == posix_errno))
            {
               /* Configuration directory is now available */
               strcat(*pathname, filename);
               res = 0;
            }
            else
            {
               print_error("FILTER: Cannot create configuration directory");
            }
         }
      }
      posix_free((void*) homedir);
   }
   else  { print_error("FILTER: Environment variable 'HOME' is not defined"); }

   return(res);
}


/* ========================================================================== */
/*! \brief Initialize filter module
 *
 * \param[in] utf8  Flag indicating that the locale use UTF-8 encoding
 *
 * \attention
 * Remember that the locale must use either UTF-8 or ISO8859-1 codeset or be the
 * POSIX locale.
 *
 * Step1 (only if \c CONF_SCORERC is configured):
 * - Rename current \c scorefile to \c scorefile.old
 * - Copy pathname configured with \c CONF_SCORERC to \c scorefile
 *
 * Step 2:
 * - Open and lock scorefile
 * - Load rules from scorefile to memory
 *
 * It is required that 'SSIZE_MAX' is at least 'INT_MAX' (must be checked by
 * build system).
 */

int  filter_init(int  utf8)
{
   int  res = -1;
   char  buf[1];
   const char*  scorerc = config[CONF_SCORERC].val.s;
   char*  scorepathname = NULL;
   char*  oldscorepathname = NULL;
   int  rv;
   struct_posix_stat  state;
   int  fd = -1;
   FILE*  fs = NULL;
   char*  data = NULL;
   size_t  len;
   char*  line = NULL;
   posix_ssize_t  readlen;
   struct filter*  rule;
#if CFG_USE_XSI
   const char*  loc_ctype;
#endif  /* CFG_USE_XSI */

#if !(CFG_USE_CLB || CFG_USE_XSI)
   print_error("FILTER: Regular expression support disabled by configuration");
#endif  /* !(CFG_USE_CLB || CFG_USE_XSI) */

   /* Set locale and check codeset */
   filter_locale = FILTER_CS_ASCII;
#if !CFG_USE_XSI
#  if CFG_USE_CLB
   print_error("FILTER: Cannot set locale due to configuration");
#  endif  /* CFG_USE_CLB */
#else  /* !CFG_USE_XSI */
   loc_ctype = posix_setlocale(POSIX_LC_CTYPE, "");
   if(NULL == loc_ctype)
   {
      print_error("FILTER: Setting locale for category 'LC_CTYPE' failed");
   }
   else
   {
      printf("%s: FILTER: Character classification locale: %s\n", CFG_NAME,
             loc_ctype);
      if(utf8)
      {
         printf("%s: FILTER: Cooked character classification codeset: "
                "UTF-8\n", CFG_NAME);
         filter_locale = FILTER_CS_UTF_8;
      }
      else
      {
         /* Check whether fallback to ISO8859-1 is possible */
         if( NULL != strstr(loc_ctype, "8859-1")
             || NULL != strstr(loc_ctype, "8859_1")
             || NULL != strstr(loc_ctype, "88591") )
         {
            printf("%s: FILTER: Cooked character classification codeset: "
                   "ISO-8859-1\n", CFG_NAME);
            filter_locale = FILTER_CS_ISO8859_1;
         }
         else
         {
            if(NULL == strstr(loc_ctype, "POSIX")
               && NULL == strstr(loc_ctype, "C"))
            {
               print_error("FILTER: Warning: Codeset of locale not supported");
            }
            else
            {
               printf("%s: FILTER: Cooked character classification codeset: "
                      "US-ASCII\n", CFG_NAME);
            }
         }
      }
   }
#endif  /* !CFG_USE_XSI */

   /* Calculate maximum length of score value strings */
   rv = posix_snprintf(buf, 1, "%d", FILTER_SCORE_MAX);
   if(0 <= rv)
   {
      score_len_max = (size_t) rv;
      rv = posix_snprintf(buf, 1, "%d", FILTER_SCORE_MIN);
      if(0 <= rv)
      {
         if((size_t) rv > score_len_max)  { score_len_max = (size_t) rv; }
         res = 0;
      }
   }
   if(res)
   {
      print_error("FILTER: Calculation of maximum score string length failed");
      return(res);
   }

   /* Step 1 */
   if(strlen(scorerc))
   {
      rv = posix_stat(scorerc, &state);
      if(rv)  { print_error("FILTER: Cannot stat scorerc file"); }
      else if(POSIX_S_ISREG(state.st_mode))
      {
         rv = filter_get_pathname(&scorepathname, scorefile_name);
         if(!rv)
         {
            /* Read scorerc file */
            rv = fu_open_file(scorerc, &fd, POSIX_O_RDWR, (posix_mode_t) 0);
            if(!rv)
            {
               rv = fu_lock_file(fd);
               if(!rv)
               {
                  rv = fu_read_whole_file(fd, &data, &len);
               }
               fu_close_file(&fd, NULL);
               if(!rv)
               {
                  oldscorepathname =
                     posix_malloc(strlen(scorepathname) + (size_t) 5);
                  if(NULL == oldscorepathname)
                  {
                     print_error("FILTER: Cannot allocate memory for pathname");
                  }
                  else
                  {
                     strcpy(oldscorepathname, scorepathname);
                     strcat(oldscorepathname, ".old");
                     rv = posix_rename(scorepathname, oldscorepathname);
                     if(rv)
                     {
                        print_error("FILTER: Renaming score file failed");
                     }
                     rv = fu_open_file(scorepathname, &fd,
                                       POSIX_O_WRONLY | POSIX_O_CREAT,
                                       FILTER_PERM);
                     if(!rv)
                     {
                        rv = fu_lock_file(fd);
                        if(!rv)
                        {
                           len = strlen(data);
                           rv = fu_write_to_filedesc(fd, data, len);
                        }
                        fu_close_file(&fd, NULL);
                     }
                  }
               }
            }
         }
      }
      if(rv)
      {
         print_error("FILTER: Importing scorerc failed,"
                     " using local scorefile");
      }
   }
   /* Release memory */
   posix_free((void*) data);
   posix_free((void*) oldscorepathname);
   posix_free((void*) scorepathname);
   scorepathname = NULL;

   /* Step 2 */
   rv = filter_get_pathname(&scorepathname, scorefile_name);
   if(!rv)
   {
      rv = posix_stat(scorepathname, &state);
      if(rv)  { print_error("FILTER: Cannot stat score file"); }
      else if(POSIX_S_ISREG(state.st_mode))
      {
         rv = fu_open_file(scorepathname, &fd, POSIX_O_RDWR, (posix_mode_t) 0);
         if(!rv)
         {
            rv = fu_lock_file(fd);
            if(!rv)
            {
               rv = fu_assign_stream(fd, &fs, "r");
               if(!rv)
               {
                  /* Load scoring rules */
                  while(1)
                  {
                     /* Read line */
                     readlen = posix_getline(&line, &len, fs);
                     if(-1 == readlen)
                     {
                        if(POSIX_ENOMEM == posix_errno)
                        {
                           print_error("FILTER: Cannot allocate memory"
                                       " for score file parser");
                        }
                        else
                        {
                           /* Check for error */
                           if(ferror(fs))
                           {
                              print_error("FILTER: Parse error in score file");
                           }
                        }
                     }
                     if(0 >= readlen)  { break; }
                     else
                     {
                        /* Extract data */
                        if('#' == line[0])  { continue; }
                        rv = filter_decode_rule(&rule, line, readlen, 0);
                        if(!rv)  { filter_add_score_rule(rule); }
                     }
                  }
                  posix_free((void*) line);
                  rv = 0;
               }
            }
         }
         fu_close_file(&fd, &fs);
      }
      if(rv)  { print_error("FILTER: Importing rules from score file failed"); }
   }
   /* Release memory */
   posix_free((void*) scorepathname);

#if CFG_USE_CLB || CFG_USE_XSI
   /* Initialize test group checking facility */
   if(!strlen(config[CONF_TESTGRP_ERE].val.s))
   {
      print_error("FILTER: No ERE for test group matching "
                  "found in config file");
      rv = -1;
   }
   else
   {
      /* Compile testgroup ERE */
      printf("%s: FILTER: Enabling test group checking facility\n", CFG_NAME);
      rv = filter_compile_ere(&testgroup_cs, &testgroup_ere,
                              config[CONF_TESTGRP_ERE].val.s);
   }
   if(rv)  { testgroup_ere = NULL; }
#endif  /* CFG_USE_CLB || CFG_USE_XSI */

   return(res);
}


/* ========================================================================== */
/*! \brief Shutdown filter module
 *
 * Step 1:
 * - Open and lock scorefile
 * - Save scoring rules in memory to score file
 * - Delete scoring rules from memory
 *
 * Step 2 (only if \c CONF_SCORERC is configured):
 * - Copy \c scorefile to the location configured with \c CONF_SCORERC
 */

void  filter_exit(void)
{
   const char*  scorerc = config[CONF_SCORERC].val.s;
   char*  scorepathname = NULL;
   char*  tmppathname = NULL;
   int  rv;
   int  fd = -1;
   FILE*  fs = NULL;
   int  fd_tmp = -1;
   FILE*  fs_tmp = NULL;
   char*  data = NULL;
   size_t  len;
   struct_posix_stat  state;
   char*  p;

   /* Step 1 */
   rv = filter_get_pathname(&scorepathname, scorefile_name);
   if(!rv)
   {
      /* Create scorefile if it does not exist */
      rv = posix_stat(scorepathname, &state);
      if(rv && POSIX_ENOENT == posix_errno)
      {
         fu_open_file(scorepathname, &fd_tmp,
                      POSIX_O_WRONLY | POSIX_O_CREAT, FILTER_PERM);
         fu_close_file(&fd_tmp, NULL);
      }
      /* Open scorefile */
      rv = posix_stat(scorepathname, &state);
      if(rv)  { print_error("FILTER: Cannot stat score file"); }
      else if(POSIX_S_ISREG(state.st_mode))
      {
         rv = fu_open_file(scorepathname, &fd, POSIX_O_RDWR, (posix_mode_t) 0);
         if(!rv)
         {
            rv = fu_lock_file(fd);
            if(!rv)
            {
               rv = fu_assign_stream(fd, &fs, "r");
               if(!rv)
               {
                  /* Open temporary file */
                  tmppathname = posix_malloc(strlen(scorepathname)
                                             + (size_t) 5);
                  if(NULL == tmppathname)
                  {
                     print_error("FILTER: Cannot allocate memory for pathname");
                  }
                  else
                  {
                     strcpy(tmppathname, scorepathname);
                     strcat(tmppathname, ".new");
                     rv = fu_open_file(tmppathname, &fd_tmp, POSIX_O_WRONLY
                                       | POSIX_O_CREAT | POSIX_O_TRUNC,
                                       FILTER_PERM);
                     /*
                      * Because we have the lock for the score file, it is
                      * allowed to assume that no other instance of the program
                      * currently use the temporary filename.
                      */
                     if(!rv)
                     {
                        rv = fu_assign_stream(fd_tmp, &fs_tmp, "w");
                        if(!rv)
                        {
                           rv = filter_export_score_rules(fs, fs_tmp);
                        }
                        /* Flush stream of temporary file*/
                        if (!rv)  { fu_sync(fd_tmp, fs_tmp); }
                        /* Rename temporary file to score file */
                        if(!rv)
                        {
                           rv = posix_rename(tmppathname, scorepathname);
                        }
                        if(rv)
                        {
                           if(tmppathname)  { fu_unlink_file(tmppathname); }
                        }
                     }
                     fu_close_file(&fd_tmp, &fs_tmp);
                  }
               }
            }
         }
         fu_close_file(&fd, &fs);
      }
      if(rv)  { print_error("FILTER: Exporting rules to score file failed"); }
   }
   filter_delete_score_rules();

   /* Step 2 */
   if(!rv && strlen(scorerc))
   {
      /* Read scorefile */
      rv = fu_open_file(scorepathname, &fd, POSIX_O_RDWR,
                        (posix_mode_t) 0);
      if(!rv)
      {
         rv = fu_lock_file(fd);
         if(!rv)
         {
            rv = fu_read_whole_file(fd, &data, &len);
         }
         fu_close_file(&fd, NULL);
         if(!rv)
         {
            /* Write scorerc file */
            p = posix_realloc(tmppathname, strlen(scorerc) + (size_t) 5);
            if(NULL == p)
            {
               print_error("FILTER: Cannot allocate memory for pathname");
            }
            else
            {
               tmppathname = p;
               strcpy(tmppathname, scorerc);
               strcat(tmppathname, ".new");
               rv = fu_open_file(tmppathname, &fd,
                                 POSIX_O_WRONLY | POSIX_O_CREAT, FILTER_PERM);
               if(!rv)
               {
                  rv = fu_lock_file(fd);
                  if(!rv)
                  {
                     len = strlen(data);
                     rv = fu_write_to_filedesc(fd, data, len);
                     if(rv)  { rv = fu_sync(fd, NULL); }
                     if(rv)
                     {
                        print_error(
                           "FILTER: Writing data to scorerc file failed");
                     }
                     else
                     {
                        rv = posix_rename(tmppathname, scorerc);
                        if(rv)
                        {
                           print_error(
                              "FILTER: Renaming new scorerc file failed");
                        }
                     }
                  }
                  fu_close_file(&fd, NULL);
               }
            }
         }
      }
      if(rv)
      {
         print_error("FILTER: Exporting score file data to scorerc failed");
      }
   }

   /* Release memory */
   posix_free((void*) data);
   posix_free((void*) tmppathname);
   posix_free((void*) scorepathname);

#if CFG_USE_CLB || CFG_USE_XSI
   /* Destroy testgroup ERE */
   if(NULL != testgroup_ere)
   {
      /* Destroy regular expression object */
      posix_regfree(testgroup_ere);
      posix_free((void*) testgroup_ere);
      testgroup_ere = NULL;
   }
#endif  /* CFG_USE_CLB || CFG_USE_XSI */

   /* Clear locale configuration */
   filter_locale = FILTER_CS_ASCII;
}


/* ========================================================================== */
/*! \brief Check for test group
 *
 * \param[in] group  Single newsgroup name (not list)
 *
 * The test group ERE from the configuration is used for matching.
 *
 * \return
 * - 1 if \e group is a test group
 * - 0 otherwise
 */

int  filter_check_testgroup(const char*  group)
{
   int  res = 0;
#if CFG_USE_CLB || CFG_USE_XSI
   int  rv = 0;
   const char*  string;
   const char*  p = NULL;
   enum enc_mime_cs  charset = ENC_CS_UNKNOWN;

   /* Check whether testgroup ERE really was compiled */
   if(NULL == testgroup_ere)
   {
      print_error("FILTER: Test group check failed (ERE not compiled)");
   }
   else
   {
      string = group;
      if(FILTER_CS_ASCII == testgroup_cs)
      {
         rv = enc_ascii_check(string);
      }
      else if(FILTER_CS_ISO8859_1 == testgroup_cs)
      {
         /* Try to convert data to ISO8859-1 */
         p = enc_convert_to_8bit(&charset, string, NULL);
         if(NULL == p)  { rv = -1; }
         else if(ENC_CS_ISO8859_1 != charset) { rv = -1; }
         else  { string = p; }
      }
      if(rv)
      {
         print_error("FILTER: Test group name cannot be checked "
                     "with current locale");
      }
      else if(!posix_regexec(testgroup_ere, string, 0, NULL, 0))
      {
         printf("%s: FILTER: Test group detected\n", CFG_NAME);
         res = 1;
      }
   }

   /* Release memory for ISO8859-1 string */
   if(NULL != p && group != p)  { enc_free((void*) p); }
#endif  /* CFG_USE_CLB || CFG_USE_XSI */

   return(res);
}


/* ========================================================================== */
/*! \brief Check for own article
 *
 * \param[in] he  Pointer to article hierarchy element
 *
 * The identity configuration is used as reference for matching.
 *
 * \return
 * - 1 if \e he corresponds to own article
 * - 0 otherwise
 */

int  filter_match_own(const struct core_hierarchy_element*  he)
{
   int  res = 0;

   if(NULL != he->header)
   {
      /*
       * This is the simple default rule
       * It matches to the identity configuration of the user.
       */
      if(!strcmp(he->header->from, config[CONF_FROM].val.s)
         || !strcmp(he->header->from, config[CONF_REPLYTO].val.s))
      {
         res = 1;
      }

      /* Hook in more sophisticated custom code here if desired */
   }

   return(res);
}


/* ========================================================================== */
/*! \brief Check for reply to own article
 *
 * \param[in] he  Pointer to article hierarchy element
 *
 * \return
 * - 1 if \e he corresponds to a reply to an own article
 * - 0 otherwise
 */

int  filter_match_reply_to_own(const struct core_hierarchy_element*  he)
{
   int  res = 0;

   if(NULL != he->parent)
   {
      res = filter_match_own(he->parent);
   }

   return(res);
}


/* ========================================================================== */
/*! \brief Get article score
 *
 * \param[in] he  Pointer to article hierarchy element
 *
 * \return
 * - Score of article
 * - 0 if no score is defined for article corresponding to \e he .
 */

int  filter_get_score(const struct core_hierarchy_element*  he)
{
   int  res = 0;
   struct filter*  rule = scores;
   const char*  data;
   size_t  i;
#if CFG_USE_CLB || CFG_USE_XSI
   int  rv;
   const char*  string;
   const char*  p = NULL;
   enum enc_mime_cs  charset = ENC_CS_UNKNOWN;
#endif  /* CFG_USE_CLB || CFG_USE_XSI */

   if(NULL != he->header)
   {
      while(NULL != rule)
      {
         data = NULL;
         switch(rule->type)
         {
            /* -------------------------------------------------------------- */
            /* Literal matching */
            case SCORE_TYPE_FROM:
            {
               if(NULL == data)  { data = he->header->from; }
               /* No break here is intended! */
            }
            case SCORE_TYPE_SUBJECT:
            {
               if(NULL == data)  { data = he->header->subject; }
               if(!strcmp(rule->string, data))
               {
                  res = filter_score_add(res, rule->value);
               }
               break;
            }
            /* -------------------------------------------------------------- */
            /* Literal matching against field element */
            case SCORE_TYPE_GROUP:
            {
               i = 0;
               /* Assignment in truth expression is intended! */
               while(NULL != (data = he->header->groups[i++]))
               {
                  if(!strcmp(rule->string, data))
                  {
                     res = filter_score_add(res, rule->value);
                  }
               }
               break;
            }
            /* -------------------------------------------------------------- */
            /* Extended regular expression matching */
            case SCORE_TYPE_FROM_ERE:
            {
               if(NULL == data)  { data = he->header->from; }
               /* No break here is intended! */
            }
            case SCORE_TYPE_SUBJECT_ERE:
            {
               if(NULL == data)  { data = he->header->subject; }
               /* No break here is intended! */
            }
            case SCORE_TYPE_MSGID_ERE:
            {
               if(NULL == data)  { data = he->header->msgid; }
#if CFG_USE_CLB || CFG_USE_XSI
               if(NULL == rule->ere)
               {
                  /*
                   * If this happens, the 'dcre' parameter of the score rule
                   * constructor was not used correctly.
                   */
                  print_error("FILTER: Regular expression not compiled (bug)");
               }
               else
               {
                  rv = 0;
                  string = data;
                  if(FILTER_CS_ASCII == rule->cs)
                  {
                     rv = enc_ascii_check(string);
                  }
                  else if(FILTER_CS_ISO8859_1 == rule->cs)
                  {
                     /* Try to convert data to ISO8859-1 */
                     p = enc_convert_to_8bit(&charset, string, NULL);
                     if(NULL == p)  { rv = -1; }
                     else if(ENC_CS_ISO8859_1 != charset) { rv = -1; }
                     else  { string = p; }
                  }
                  if(!rv && !posix_regexec(rule->ere, string, 0, NULL, 0))
                  {
                     res = filter_score_add(res, rule->value);
                  }
                  /* Release memory for ISO8859-1 string */
                  if(NULL != p && data != p)  { enc_free((void*) p); }
               }
#endif  /* CFG_USE_CLB || CFG_USE_XSI */
               break;
            }
            /* -------------------------------------------------------------- */
            default:
            {
               print_error("FILTER: Unknown type of score rule (bug)");
               break;
            }
         }
         rule = rule->next;
      }
   }

   return(res);
}


/* @} */

/* EOF */
