/* ========================================================================== */
/*! \file
 * \brief Keyed-Hash Message Authentication Code
 *
 * Copyright (c) 2012-2015 by the developers. See the LICENSE file for details.
 *
 * If nothing else is specified, function return zero to indicate success
 * and a negative value to indicate an error.
 */


/* ========================================================================== */
/* Include headers */

#include "posix.h"  /* Include this first because of feature test macros */

#include "config.h"

#if CFG_USE_TLS  /* This requires data from "config.h" */
#  include <openssl/evp.h>
#  include <openssl/hmac.h>
#endif  /* CFG_USE_TLS */
#include <string.h>

#include "hmac.h"
#include "main.h"


/* ========================================================================== */
/*! \addtogroup HMAC
 *
 * This module provides optional support for Keyed-Hash Message Authentication
 * Codes. It is intended to comply with RFC2104.
 *
 * \attention
 * This module currently works only if TLS support is available (because both
 * share the cryptographic functions provided by OpenSSL).
 */
/* @{ */


/* ========================================================================== */
/*! \brief Message Authentication Code based on SHA-1 hash algorithm
 *
 * \param[in]  text      Message
 * \param[in]  text_len  Message length
 * \param[in]  key       Key
 * \param[in]  key_len   Key length
 * \param[out] md        Result
 *
 * The values of paramaters \e text_len and \e key_len must represent bytes.
 *
 * On success, the result is written to \e md and has a length of
 * \ref HMAC_SHA1_LEN bytes.
 *
 * \return
 * - 0 on success
 * - -1 on error
 */

int  hmac_sha1(const char*  text, size_t  text_len,
               const char*  key, size_t  key_len, unsigned char*  md)
{
   int  res = -1;
#if CFG_USE_TLS
   HMAC_CTX*  ctx;
   unsigned int  md_len = 0;

   /* Create context */
   ctx = (HMAC_CTX*) posix_malloc(sizeof(HMAC_CTX));
   if(NULL == ctx)  { print_error("HMAC: Creating context failed"); }
   /* Check whether length values can be represented as data type 'int' */
   else if(!(POSIX_INT_MAX >= text_len && POSIX_INT_MAX >= key_len))
   {
      print_error("HMAC: Message or key is too long");
   }
   else
   {
      /* Delegate job to OpenSSL */
      HMAC_CTX_init(ctx);
      md = HMAC(EVP_sha1(), (void*) key, (int) key_len, (unsigned char*) text,
                (int) text_len, (unsigned char*) md, &md_len);
      if(NULL != md && HMAC_SHA1_LEN == (size_t) md_len)  { res = 0; }
      HMAC_CTX_cleanup(ctx);
   }
   posix_free((void*) ctx);
#endif  /* CFG_USE_TLS */

   return(res);
}


/* ========================================================================== */
/*! \brief Initialize HMAC module */

void  hmac_init(void)
{
   return;
}


/* ========================================================================== */
/*! \brief Shutdown HMAC module */

void  hmac_exit(void)
{
   return;
}


/* @} */

/* EOF */
