/* ========================================================================== */
/*! \file
 * \brief Internet connection handling
 *
 * Copyright (c) 2015 by the developers. See the LICENSE file for details.
 *
 * If nothing else is specified, functions return zero to indicate success
 * and a negative value to indicate an error.
 */


/* ========================================================================== */
/* Include headers */

#include "posix.h"  /* Include this first because of feature test macros */

#include <string.h>

#include "config.h"
#include "inet.h"
#include "main.h"


/* ========================================================================== */
/*! \addtogroup INET
 *
 * Internet name resolution, socket and connection handling.
 *
 * \todo IPv6 support (POSIX module already has IPv6 support).
 */
/* @{ */


/* ========================================================================== */
/*! \brief Name resolver
 *
 * \param[in]  host     Name of host
 * \param[in]  service  Name of service
 * \param[out] sa       Pointer to server address
 *
 * \note
 * The value at the location pointed to by \e sa is undefined after error
 * status is returned.
 */

int  inet_name_resolver(const char*  host, const char*  service,
                        struct_posix_sockaddr_in*  sa)
{
   static const char*  prefix = "INET: ";
   int  res = 0;
   struct_posix_addrinfo  hints;
   struct_posix_addrinfo*  sai;
   int  rv;
   const char*  err_str;
   char*  ebuf = NULL;
   size_t  len;

   memset((void*) &hints, 0, sizeof(hints));
   /* Only return results for address families configured on local system */
   hints.ai_flags = POSIX_AI_ADDRCONFIG;
#if CFG_USE_IP6 && 0
   printf("%s: INET: IPv6 mode of name resolver enabled\n", CFG_NAME);
   hints.ai_family = POSIX_AF_UNSPEC;  /* Results for any address family */
#else  /* CFG_USE_IP6 */
   hints.ai_family = POSIX_AF_INET;    /* Results for IPv4 only */
#endif  /* CFG_USE_IP6 */
   hints.ai_socktype = POSIX_SOCK_STREAM;
   hints.ai_protocol = 0;
   rv = posix_getaddrinfo(host, service, &hints, &sai);
   if(rv)
   {
      print_error("INET: getaddrinfo() failed");
      err_str = posix_gai_strerror(rv);
      if(NULL != err_str)
      {
         len = strlen(prefix);
         len += strlen(err_str);
         ebuf = (char*) posix_malloc(++len);
         if(NULL != ebuf)
         {
            strcpy(ebuf, prefix);
            strcat(ebuf, err_str);
            print_error(ebuf);
         }
         posix_free((void*) ebuf);
      }
      res = -1;
   }
   else
   {
      memcpy((void*) sa, (void*) sai->ai_addr,
             sizeof(struct_posix_sockaddr_in));
      posix_freeaddrinfo(sai);
   }

   return(res);
}


/* ========================================================================== */
/*! \brief Create socket
 *
 * \param[out] sd  Pointer to socket descriptor
 * \param[in]  af  Address family
 *
 * The value for \e af must be \c POSIX_INET or \c POSIX_INET6 respectively.
 *
 * \note
 * The value -1 is written to the location pointed to by \e sd on error.
 */

int  inet_socket_create(int*  sd, int  af)
{
   int  res = 0;
   int  rv;

#if CFG_USE_IP6
   if(POSIX_AF_INET6 == af)
   {
      printf("%s: INET: Creating IPv6 socket\n", CFG_NAME);
   }
#endif  /* CFG_USE_IP6 */

   *sd = posix_socket(af, POSIX_SOCK_STREAM, 0);
   if(0 > *sd)
   {
      print_error("INET: socket() failed");
      *sd = -1;
      res = -1;
   }
   else
   {
      /* Set "Close on exec()" flag */
      do  { rv = posix_fcntl(*sd, POSIX_F_SETFD, POSIX_FD_CLOEXEC); }
      while(-1 == rv && POSIX_EINTR == posix_errno);
      if(-1 == rv)
      {
         print_error("INET: fcntl() failed");
         inet_close(sd);
         res = -1;
      }
   }

   return(res);
}


/* ========================================================================== */
/*! \brief Establish stream oriented connection
 *
 * \param[in] sd  Socket descriptor
 * \param[in] sa  Pointer to server address
 */

int  inet_connect(int  sd, struct_posix_sockaddr_in*  sa)
{
   int  res = 0;
   int  rv;
   posix_socklen_t  sa__len;  /* Note: IRIX use 'sa_len' internally */
   struct_posix_pollfd  fds[1];
   posix_socklen_t  opt_len;
   int  refused = 0;

   sa__len = sizeof(struct_posix_sockaddr_in);
   /*
    * If the compiler complain about the typecast:
    * This is the intended way to use the BSD/POSIX socket API.
    * The implementation of 'posix_connect()' is responsible to convert the
    * pointer back to the correct type before dereferencing it.
    * For GCC you can use the "-fno-strict-aliasing" option as workaround
    * (if this should not be the case).
    */
   rv = posix_connect(sd, (struct_posix_sockaddr*) sa, sa__len);
   if(rv)
   {
      if(POSIX_ECONNREFUSED == posix_errno)  { refused = 1; }
      res = -1;
   }
   if(-1 == rv && POSIX_EINTR == posix_errno)
   {
      /* Interrupted by signal => Connection is established asynchronously */
      do
      {
         fds[0].fd = sd;
         fds[0].events = POSIX_POLLIN;
         fds[0].revents = 0;
         rv = posix_poll(fds, 1, -1);
      }
      while(-1 == rv && POSIX_EINTR == posix_errno);
      if(1 == rv)
      {
         /* Get return value of 'connect()' */
         opt_len = sizeof(int);
         rv = posix_getsockopt(sd, POSIX_SOL_SOCKET, POSIX_SO_ERROR,
                               &res, &opt_len);
         if(-1 == rv)  { res = -1; }
      }
   }

   /* Check for error */
   if(res)
   {
      if(refused)  { print_error("INET: Connection refused"); }
      else  { print_error("INET: connect() failed"); }
   }

   return(res);
}


/* ========================================================================== */
/*! \brief Close connection and destroy socket
 *
 * \param[in,out] sd  Pointer to socket descriptor
 *
 * \note
 * It is save to call this funktion with \e sd pointing to \c -1 . This will
 * execute as NOP.
 *
 * \note
 * The value -1 is present at the location pointed to by \e sd in any case after
 * return.
 */

void  inet_close(int*  sd)
{
   if(-1 != *sd)
   {
      posix_close(*sd);
      *sd = -1;
   }
}


/* @} */

/* EOF */
