/* ========================================================================== */
/*! \file
 * \brief Test of \c enc_convert_to_utf8_nfc() implementation
 *
 * Copyright (c) 2012-2015 by the developers. See the LICENSE file for details.
 */


/* ========================================================================== */
/* Include headers */

#include "posix.h"  /* Include this first because of feature test macros */

#include <stdio.h>
#include <string.h>

#include "config.h"

#include "encoding.h"
#include "test.h"
#include "test_unicode.h"


/* ========================================================================== */
/*! \addtogroup TEST */
/* @{ */


/* ========================================================================== */
/*! \brief Test \c enc_convert_to_utf8_nfc() implementation
 *
 * \note
 * The UTF-7 transformation format use base64 encoded UTF-16BE as internal
 * representation. Therefore all Unicode codepoints beyond the BMP must be
 * encoded using surrogate codepoints (that are forbidden in UTF-8).
 *
 * The following cases are tested:
 * - ASCII only (trivial)
 * - Unicode already in NFC normalization (NFC quick check)
 * - Unicode precomposed but with composition exception (requires lookup table)
 * - Unicode with NFD normalization (trivial canonical composition)
 * - Unicode with noncanonical order A (canonical reordering and composition)
 * - Unicode with noncanonical order B (canonical reordering)
 * - Unicode singleton (decomposition to another single codepoint)
 * - UTF-7 to UTF-8 conversion (and conversion from NFD to NFC normalization)
 * - UTF-7 to UTF-8 conversion (with codepoint that require surrogate pair)
 *
 * \return
 * - \c EXIT_SUCCESS on success
 * - \c EXIT_FAILURE on error
 */

int  test_unicode(void)
{
#define TS_NUM  (size_t) 9  /* Number of test strings */
#define TS_UTF7  (size_t) 7  /* First index of UTF-7 section */
   static const char*  ts[TS_NUM] =
   {
      /* UTF-8 section */
      "This is an ASCII string",
      "This is an Unicode string: \xC3\xA4word",
      "This is an Unicode string: \xE0\xAD\x9Cword",
      "This is an Unicode string: a\xCC\x88word",
      "This is an Unicode string: start\xCE\xB1\xCC\x94\xCC\x81\xCD\x85word",
      "This is an Unicode string: start\xCE\xB1\xCC\x81\xCC\x94\xCD\x85word",
      "This is an Unicode string: \xE2\x84\xA6word",
      /* UTF-7 section */
      "This is an Unicode string: hundertf+AHUDCA-nfzig",
      "This is an Unicode string: Violinschl+APw-ssel (+2DTdHg-)"
   };
   static const char*  rs[TS_NUM] =  {
      /* UTF-8 section */
      "This is an ASCII string",
      "This is an Unicode string: \xC3\xA4word",
      "This is an Unicode string: \xE0\xAC\xA1\xE0\xAC\xBCword",
      "This is an Unicode string: \xC3\xA4word",
      "This is an Unicode string: start\xE1\xBE\x85word",
      "This is an Unicode string: start\xE1\xBE\xB4\xCC\x94word",
      "This is an Unicode string: \xCE\xA9word",
      /* UTF-7 section */
      "This is an Unicode string: hundertf\xC3\xBCnfzig",
      "This is an Unicode string: Violinschl\xC3\xBCssel (\xF0\x9D\x84\x9E)"
   };
   int  res = POSIX_EXIT_SUCCESS;
   size_t  i;
   const char*  buf;
   const char*  loc;

   for(i = 0; i < TS_NUM; ++i)
   {
      if(TS_UTF7 <= i)
      {
         buf = enc_convert_to_utf8_nfc(ENC_CS_UTF_7, ts[i]);
      }
      else
      {
         buf = enc_convert_to_utf8_nfc(ENC_CS_UTF_8, ts[i]);
      }
      if(NULL == buf)
      {
         res = POSIX_EXIT_FAILURE;
         break;
      }
      if(strcmp(rs[i], buf))
      {
         print_error("Result is not correct");
         /* For debugging */
#if CFG_USE_XSI && !CFG_NLS_DISABLE
         loc = posix_setlocale(POSIX_LC_CTYPE, "");
#else  /* CFG_USE_XSI && !CFG_NLS_DISABLE */
         loc = NULL;
#endif  /* CFG_USE_XSI && !CFG_NLS_DISABLE */
         if(NULL == loc)
         {
            print_error("Setting locale for debug messages failed");
         }
         else
         {
            /* Print Unicode data only if terminal use Unicode locale */
            if(NULL == strstr(loc, "UTF") && NULL == strstr(loc, "utf"))
            {
               print_error(
                  "Debug messages can't be printed with current locale");
            }
            else
            {
               fprintf(stderr, TEST_TAB "Input data: \"%s\"\n", ts[i]);
               fprintf(stderr, TEST_TAB "Result is : \"%s\"\n", buf);
               fprintf(stderr, TEST_TAB "Should be : \"%s\"\n", rs[i]);
            }
         }
         res = POSIX_EXIT_FAILURE;
         break;
      }
      if(buf != ts[i])  { enc_free((void*) buf); }
   }

   return(res);
}


/* @} */

/* EOF */
